# """
# Hönl-London factors for doublets, formulas from Kovacs 1969 p130
#
# References:
#     [1] Istvan Kovacs, Rotational Structure in the spectra of diatomic molecules.
#         American Elsevier, 1969
#
# """
#
# from collections import OrderedDict
#
# __all__ = ["get_honllondon_formulas", "quanta_to_branch"]
#
#
#
#
# def quanta_to_branch(Jl, J2l, spin):
#     """
#     Singlet only has branches P/Q/R
#
#     Args:
#         Jl: J upper or J'
#         J2l: J lower or  J''
#
#     Returns: branch letter: "P"/"Q"/"R"
#
#     Example:
#
#     >>> quanta_to_branch(10.5, 11.5, 1)
#     'P1'
#     >>> quanta_to_branch(11.5, 10.5, 2)
#     'R2'
#     >>> quanta_to_branch(10.5, 10.5, 1)
#     'Q1'
#
#     """
#     if Jl > J2l:
#         br = "R"
#     elif Jl == J2l:
#         br =  "Q"
#     else:
#         br = "P"
#     ret = "{}{:1d}".format(br, spin)
#     return ret
#
#
# def get_honllondon_formulas(LAML, LAM2L):
#     """Function factory: returns (P1(J), Q1(J), ...)
#
#     Args:
#         LAML: Lambda sup:
#           0 for Sigma
#           1 for Pi
#           2 for Delta
#           3 for Phi
#         LAM2L: Lambda inf (0/1/2/3 as above)
#
#     Returns:
#         dictionary with keys (P1, Q1, R1, P21, Q21, R21, P12, Q12, R12, P2, Q2, R2)
#
#     all functions are functions of J, i.e., P1 = P1(J) (they are adjusted for specific LAML, LAM2L)
#
#     Order as in Kovacs 1969, p130, 1st column
#
#     Examples:
#
#     >>> formulas = get_honllondon_formulas(1, 0)
#
#     >>> formulas = get_honllondon_formulas(0, 1)
#
#     >>> J = 10.5
#     >>> formulas = get_honllondon_formulas(1, 0)  # delta Lambda = +1
#     >>> p1, q1, r1, p21, q21, r21, p12, q12, r12, p2, q2, r2 = [f(J) for f in formulas.values()]
#
#     >>> J = 10.5
#     >>> formulas = get_honllondon_formulas(1, 0)  # delta Lambda = +1
#     >>> [f(J) for f in formulas.values()]
#     [4.714285714285714, 10.952380952380953, 6.260869565217392, 0.047619047619047616, 0.02484472049689441, 0.0, 0.0, 0.020703933747412008, 0.043478260869565216, 4.761904761904762, 10.956521739130435, 6.217391304347826]
#
#     >>> S, DELTAK = 0.5, 0  # spin, delta Kronecker
#     >>> J = 3
#     >>> factor = 2/((2*J+1)*(2*S+1)*(2-DELTAK))
#     >>> normalized = [f(J)*factor for f in get_honllondon_formulas(0, 1).values()]
#     >>> normalized
#     [0.10714285714285715, 0.24891774891774893, 0.1422924901185771, 0.0010822510822510823, 0.0005646527385657821, 0.0, 0.0, 0.0004705439488048184, 0.0009881422924901185, 0.10822510822510822, 0.24901185770750991, 0.14130434782608697]
#     >>> sum(normalized)
#     1.0
#
#     >>> S, DELTAK = 0.5, 0  # spin, delta Kronecker
#     >>> J = 1.5
#     >>> factor = 2/((2*J+1)*(2*S+1)*(2-DELTAK))
#     >>> normalized = [f(J)*factor for f in get_honllondon_formulas(0, 1).values()]
#     >>> sum(normalized)
#     1.0
#     """
#
#     if LAML > LAM2L:
#         ret = OrderedDict((("P1", lambda J: formula0(J, LAML, LAM2L)),
#                            ("Q1", lambda J: formula1(J, LAML, LAM2L)),
#                            ("R1", lambda J: formula2(J, LAML, LAM2L)),
#                            ("P21", lambda J: formula3(J, LAML, LAM2L)),
#                            ("Q21", lambda J: formula4(J, LAML, LAM2L)),
#                            ("R21", lambda J: formula5(J, LAML, LAM2L)),
#                            ("P12", lambda J: formula6(J, LAML, LAM2L)),
#                            ("Q12", lambda J: formula7(J, LAML, LAM2L)),
#                            ("R12", lambda J: formula8(J, LAML, LAM2L)),
#                            ("P2", lambda J: formula9(J, LAML, LAM2L)),
#                            ("Q2", lambda J: formula10(J, LAML, LAM2L)),
#                            ("R2", lambda J: formula11(J, LAML, LAM2L)),
#                          ))
#     elif LAML < LAM2L:
#         ret = OrderedDict((("P1", lambda J: formula2(J-1, LAML, LAM2L)),
#                            ("Q1", lambda J: formula1(J, LAML, LAM2L)),
#                            ("R1", lambda J: formula0(J+1, LAML, LAM2L)),
#                            ("P21", lambda J: formula8(J-1, LAML, LAM2L)),
#                            ("Q21", lambda J: formula7(J, LAML, LAM2L)),
#                            ("R21", lambda J: formula6(J+1, LAML, LAM2L)),
#                            ("P12", lambda J: formula5(J-1, LAML, LAM2L)),
#                            ("Q12", lambda J: formula4(J, LAML, LAM2L)),
#                            ("R12", lambda J: formula3(J+1, LAML, LAM2L)),
#                            ("P2", lambda J: formula11(J-1, LAML, LAM2L)),
#                            ("Q2", lambda J: formula10(J, LAML, LAM2L)),
#                            ("R2", lambda J: formula9(J+1, LAML, LAM2L)),
#                          ))
#     else:
#         raise ValueError("Invalid delta Lambda")
#
#     return ret
#
#
# def UPLUS(J, LAM):
#     """Formula (3), p. 128"""
#     return 2. * (J - LAM + 0.5)
#
# def UMINUS(J, LAM):
#     """Formula (3), p. 128"""
#     return 2. * (J + LAM + 0.5)
#
# def CPLUS(J, LAM):
#     """Formula (3), p. 128"""
#     return 4. * (J + 0.5) * (J - LAM + 0.5)
#
# def CMINUS(J, LAM):
#     """Formula (3), p. 128"""
#     return 4. * (J + 0.5) * (J + LAM + 0.5)
#
#
# def formula0(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for P1(J) or R1(J-1)"""
#     return (((J-LAM2L-1.5)*(J-LAM2L-0.5))/(8*J*CMINUS(J-1, LAML)*
#     CMINUS(J, LAM2L)))*((UMINUS(J-1, LAML)*UMINUS(J, LAM2L) +
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
# def formula1(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for Q1(J)"""
#     return (((J-LAM2L-0.5)*(J+0.5)*(J+LAM2L+1.5))/(4*J*(J+1)*
#     CMINUS(J, LAML)*CMINUS(J, LAM2L)))*((UMINUS(J, LAML)*UMINUS(J, LAM2L) +
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula2(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for R1(J) or P1(J+1)"""
#     return (((J+LAM2L+1.5)*(J+LAM2L+2.5))/(8*(J+1)*CMINUS(J+1, LAML)*
#     CMINUS(J, LAM2L)))*((UMINUS(J+1, LAML)*UMINUS(J, LAM2L) +
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#      #   P1(J+1)=(((J+LAM2L+1.5)*(J+LAM2L+2.5))/(8*(J+1)*CMENOSL(J+1)*
#      # 1 CMENOS2L(J)))*((UMENOSL(J+1)*UMENOS2L(J) +
#      # 2 4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula3(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for P21(J) or R12(J-1)"""
#     return (((J-LAM2L-1.5)*(J-LAM2L-0.5))/(8*J*CPLUS(J-1, LAML)*
#     CMINUS(J, LAM2L)))*((UPLUS(J-1, LAML)*UMINUS(J, LAM2L) -
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula4(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for Q21(J) or Q12(J)"""
#     return (((J-LAM2L-0.5)*(J+0.5)*(J+LAM2L+1.5))/(4*J*(J+1)*
#     CPLUS(J, LAML)*CMINUS(J, LAM2L)))*((UPLUS(J, LAML)*UMINUS(J, LAM2L) -
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
# def formula5(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for R21(J) or P12(J+1)"""
#     return (((J+LAM2L+1.5)*(J+LAM2L+2.5))/(8*(J+1)*CPLUS(J+1, LAML)*
#     CMINUS(J, LAM2L)))*((UPLUS(J+1, LAML)*UMINUS(J, LAM2L) -
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula6(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for P21(J) or R21(J-1)"""
#     return (((J-LAM2L-1.5)*(J-LAM2L-0.5))/(8*J*CMINUS(J-1, LAML)*
#     CPLUS(J, LAM2L)))*((UMINUS(J-1, LAML)*UPLUS(J, LAM2L) -
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula7(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for Q12(J) or Q21(J)"""
#     return (((J-LAM2L-0.5)*(J+0.5)*(J+LAM2L+1.5))/(4*J*(J+1)*
#     CMINUS(J, LAML)*CPLUS(J, LAM2L)))*((UMINUS(J, LAML)*UPLUS(J, LAM2L) -
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula8(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for R12(J) or P21(J+1)"""
#     return (((J+LAM2L+1.5)*(J+LAM2L+2.5))/(8*(J+1)*CMINUS(J+1, LAML)*
#     CPLUS(J, LAM2L)))*((UMINUS(J+1, LAML)*UPLUS(J, LAM2L) -
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula9(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for P2(J) or R2(J-1)"""
#     return (((J-LAM2L-1.5)*(J-LAM2L-0.5))/(8*J*CPLUS(J-1, LAML)*
#     CPLUS(J, LAM2L)))*((UPLUS(J-1, LAML)*UPLUS(J, LAM2L) +
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula10(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for Q2(J)"""
#     return (((J-LAM2L-0.5)*(J+0.5)*(J+LAM2L+1.5))/(4*J*(J+1)*
#     CPLUS(J, LAML)*CPLUS(J, LAM2L)))*((UPLUS(J, LAML)*UPLUS(J, LAM2L) +
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)
#
#
# def formula11(J, LAML, LAM2L):
#     """Table 3.7 (p. 130): Formula for R2(J) or P2(J+1)"""
#     return (((J+LAM2L+1.5)*(J+LAM2L+2.5))/(8*(J+1)*CPLUS(J+1, LAML)*
#     CPLUS(J, LAM2L)))*((UPLUS(J+1, LAML)*UPLUS(J, LAM2L) +
#     4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)






"""
Hönl-London factors for doublets, formulas from Kovacs 1969 p130
"""

from collections import OrderedDict

__all__ = ["get_honllondon_formulas", "quanta_to_branch"]




def quanta_to_branch(Jl, J2l, spin):
    """
    Singlet only has branches P/Q/R

    Args:
        Jl: J upper or J'
        J2l: J lower or  J''

    Returns: branch letter: "P"/"Q"/"R"

    >>> quanta_to_branch(10.5, 11.5, 1)
    'P1'
    >>> quanta_to_branch(11.5, 10.5, 2)
    'R2'
    >>> quanta_to_branch(10.5, 10.5, 1)
    'Q1'

    """
    if Jl > J2l:
        br = "R"
    elif Jl == J2l:
        br =  "Q"
    else:
        br = "P"
    ret = "{}{:1d}".format(br, spin)
    return ret


def get_honllondon_formulas(LAML, LAM2L):
    """Function factory: returns (P1(J), Q1(J), ...)

    Args:
        LAML: Lambda sup:
          0 for Sigma
          1 for Pi
          2 for Delta
          3 for Phi
        LAM2L: Lambda inf (0/1/2/3 as above)

    Returns:
        dictionary with keys (P1, Q1, R1, P21, Q21, R21, P12, Q12, R12, P2, Q2, R2)

    all functions are functions of J, i.e., P1 = P1(J) (they are adjusted for specific LAML, LAM2L)

    Order as in Kovacs 1969, p130, 1st column

    Examples:

    >>> formulas = get_honllondon_formulas(1, 0)

    >>> formulas = get_honllondon_formulas(0, 1)

    >>> J = 10.5
    >>> formulas = get_honllondon_formulas(1, 0)  # delta Lambda = +1
    >>> p1, q1, r1, p21, q21, r21, p12, q12, r12, p2, q2, r2 = [f(J) for f in formulas.values()]

    >>> J = 10.5
    >>> formulas = get_honllondon_formulas(1, 0)  # delta Lambda = +1
    >>> [f(J) for f in formulas.values()]
    [4.714285714285714, 10.952380952380953, 6.260869565217392, 0.047619047619047616, 0.02484472049689441, 0.0, 0.0, 0.020703933747412008, 0.043478260869565216, 4.761904761904762, 10.956521739130435, 6.217391304347826]


    **Sanity check examples**. In the following examples, the sum of all "normalized" Hönl-London
    factors (HLF) for a given J must be 1 ("higher-order" branches are not considered because their
    HLFs are negligible.

    >>> S, DELTAK = 0.5, 0  # spin, delta Kronecker
    >>> J = 3
    >>> factor = 2/((2*J+1)*(2*S+1)*(2-DELTAK))
    >>> normalized = [f(J)*factor for f in get_honllondon_formulas(0, 1).values()]
    >>> normalized
    [0.14583333333333334, 0.23809523809523808, 0.10044642857142856, 0.011904761904761906, 0.003720238095238095, 0.0, 0.0, 0.006696428571428571, 0.008928571428571428, 0.13392857142857142, 0.24107142857142855, 0.109375]
    >>> sum(normalized)
    0.9999999999999998


    >>> S, DELTAK = 0.5, 0  # spin, delta Kronecker
    >>> J = 1.5
    >>> factor = 2/((2*J+1)*(2*S+1)*(2-DELTAK))
    >>> normalized = [f(J)*factor for f in get_honllondon_formulas(0, 1).values()]
    >>> sum(normalized)
    1.0
    """

    if LAML > LAM2L:
        ret = OrderedDict((("P1", lambda J: formula0(J, LAML, LAM2L)),
                           ("Q1", lambda J: formula1(J, LAML, LAM2L)),
                           ("R1", lambda J: formula2(J, LAML, LAM2L)),
                           ("P21", lambda J: formula3(J, LAML, LAM2L)),
                           ("Q21", lambda J: formula4(J, LAML, LAM2L)),
                           ("R21", lambda J: formula5(J, LAML, LAM2L)),
                           ("P12", lambda J: formula6(J, LAML, LAM2L)),
                           ("Q12", lambda J: formula7(J, LAML, LAM2L)),
                           ("R12", lambda J: formula8(J, LAML, LAM2L)),
                           ("P2", lambda J: formula9(J, LAML, LAM2L)),
                           ("Q2", lambda J: formula10(J, LAML, LAM2L)),
                           ("R2", lambda J: formula11(J, LAML, LAM2L)),
                         ))
    elif LAML < LAM2L:
        ret = OrderedDict((("P1", lambda J: formula2(J-1, LAM2L, LAML)),
                           ("Q1", lambda J: formula1(J, LAM2L, LAML)),
                           ("R1", lambda J: formula0(J+1, LAM2L, LAML)),
                           ("P21", lambda J: formula8(J-1, LAM2L, LAML)),
                           ("Q21", lambda J: formula7(J, LAM2L, LAML)),
                           ("R21", lambda J: formula6(J+1, LAM2L, LAML)),
                           ("P12", lambda J: formula5(J-1, LAM2L, LAML)),
                           ("Q12", lambda J: formula4(J, LAM2L, LAML)),
                           ("R12", lambda J: formula3(J+1, LAM2L, LAML)),
                           ("P2", lambda J: formula11(J-1, LAM2L, LAML)),
                           ("Q2", lambda J: formula10(J, LAM2L, LAML)),
                           ("R2", lambda J: formula9(J+1, LAM2L, LAML)),
                         ))
    else:
        raise ValueError("Invalid delta Lambda")

    return ret


def UPLUSL(J, LBIG):
    return 2. * (J - LBIG + 0.5)


def UMINUSL(J, LBIG):
    return 2. * (J + LBIG + 0.5)


def CPLUSL(J, LBIG):
    return 4. * (J + 0.5) * (J - LBIG + 0.5)


def CMINUSL(J, LBIG):
    return 4. * (J + 0.5) * (J + LBIG + 0.5)


def UPLUS2L(J, LSMALL):
    return 2.*(J-LSMALL+0.5)


def UMINUS2L(J, LSMALL):
    return 2.*(J+LSMALL+0.5)


def CPLUS2L(J, LSMALL):
    return 4.*(J+0.5)*(J-LSMALL+0.5)


def CMINUS2L(J, LSMALL):
    return 4.*(J+0.5)*(J+LSMALL+0.5)


def formula0(J, LBIG, LSMALL):
    """Formula for P1(J) or R1(J-1)"""
    return (((J-LSMALL-1.5)*(J-LSMALL-0.5))/(8*J*CMINUSL(J-1, LBIG)*
    CMINUS2L(J, LSMALL)))*((UMINUSL(J-1, LBIG)*UMINUS2L(J, LSMALL) +
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)

def formula1(J, LBIG, LSMALL):
    """Formula for Q1(J)"""
    return (((J-LSMALL-0.5)*(J+0.5)*(J+LSMALL+1.5))/(4*J*(J+1)*
    CMINUSL(J, LBIG)*CMINUS2L(J, LSMALL)))*((UMINUSL(J, LBIG)*UMINUS2L(J, LSMALL) +
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula2(J, LBIG, LSMALL):
    """Formula for R1(J) or P1(J+1)"""
    return (((J+LSMALL+1.5)*(J+LSMALL+2.5))/(8*(J+1)*CMINUSL(J+1, LBIG)*
    CMINUS2L(J, LSMALL)))*((UMINUSL(J+1, LBIG)*UMINUS2L(J, LSMALL) +
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)

     #   P1(J+1)=(((J+LAM2L+1.5)*(J+LAM2L+2.5))/(8*(J+1)*CMENOSL(J+1)*
     # 1 CMENOS2L(J)))*((UMENOSL(J+1)*UMENOS2L(J) +
     # 2 4*(J-LAM2L+0.5)*(J+LAM2L+0.5))**2.)


def formula3(J, LBIG, LSMALL):
    """Formula for P21(J) or R12(J-1)"""
    return (((J-LSMALL-1.5)*(J-LSMALL-0.5))/(8*J*CPLUSL(J-1, LBIG)*
    CMINUS2L(J, LSMALL)))*((UPLUSL(J-1, LBIG)*UMINUS2L(J, LSMALL) -
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula4(J, LBIG, LSMALL):
    """Formula for Q21(J) or Q12(J)"""
    return (((J-LSMALL-0.5)*(J+0.5)*(J+LSMALL+1.5))/(4*J*(J+1)*
    CPLUSL(J, LBIG)*CMINUS2L(J, LSMALL)))*((UPLUSL(J, LBIG)*UMINUS2L(J, LSMALL) -
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)

def formula5(J, LBIG, LSMALL):
    """Formula for R21(J) or P12(J+1)"""
    return (((J+LSMALL+1.5)*(J+LSMALL+2.5))/(8*(J+1)*CPLUSL(J+1, LBIG)*
    CMINUS2L(J, LSMALL)))*((UPLUSL(J+1, LBIG)*UMINUS2L(J, LSMALL) -
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula6(J, LBIG, LSMALL):
    """Formula for P21(J) or R21(J-1)"""
    return (((J-LSMALL-1.5)*(J-LSMALL-0.5))/(8*J*CMINUSL(J-1, LBIG)*
    CPLUS2L(J, LSMALL)))*((UMINUSL(J-1, LBIG)*UPLUS2L(J, LSMALL) -
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula7(J, LBIG, LSMALL):
    """Formula for Q12(J) or Q21(J)"""
    return (((J-LSMALL-0.5)*(J+0.5)*(J+LSMALL+1.5))/(4*J*(J+1)*
    CMINUSL(J, LBIG)*CPLUS2L(J, LSMALL)))*((UMINUSL(J, LBIG)*UPLUS2L(J, LSMALL) -
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula8(J, LBIG, LSMALL):
    """Formula for R12(J) or P21(J+1)"""
    return (((J+LSMALL+1.5)*(J+LSMALL+2.5))/(8*(J+1)*CMINUSL(J+1, LBIG)*
    CPLUS2L(J, LSMALL)))*((UMINUSL(J+1, LBIG)*UPLUS2L(J, LSMALL) -
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula9(J, LBIG, LSMALL):
    """Formula for P2(J) or R2(J-1)"""
    return (((J-LSMALL-1.5)*(J-LSMALL-0.5))/(8*J*CPLUSL(J-1, LBIG)*
    CPLUS2L(J, LSMALL)))*((UPLUSL(J-1, LBIG)*UPLUS2L(J, LSMALL) +
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula10(J, LBIG, LSMALL):
    """Formula for Q2(J)"""
    return (((J-LSMALL-0.5)*(J+0.5)*(J+LSMALL+1.5))/(4*J*(J+1)*
    CPLUSL(J, LBIG)*CPLUS2L(J, LSMALL)))*((UPLUSL(J, LBIG)*UPLUS2L(J, LSMALL) +
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)


def formula11(J, LBIG, LSMALL):
    """Formula for R2(J) or P2(J+1)"""
    return (((J+LSMALL+1.5)*(J+LSMALL+2.5))/(8*(J+1)*CPLUSL(J+1, LBIG)*
    CPLUS2L(J, LSMALL)))*((UPLUSL(J+1, LBIG)*UPLUS2L(J, LSMALL) +
    4*(J-LSMALL+0.5)*(J+LSMALL+0.5))**2.)



__all__ = ["create_simulation_report", "create_index"]


import os
from astropy.io import fits
from matplotlib import pyplot as plt
from .util import *
import numpy as np
import traceback
import glob
import a99
from .. import aosss as ao
import f311.filetypes as ft
import f311.explorer as ex


def create_index(dir_="."):
    """[re]created 'index.html' of report*"""
    ff = glob.glob(os.path.join(dir_, "report-*.html"))
    ff.sort()

    with open(os.path.join(dir_, "index.html"), "w") as html:
        html.write(_head("List of Reports"))
        html.write("<body>")
        for f in ff:
            f = os.path.basename(f)
            html.write('<p><a href="{0}">{0}</a>'.format(f))
        html.write("</body></html>")


# # HTML fragment generation routines
# Block routines (e.g., head, h1) result should always end with a "\n".
# Inline tags, NOT.


def _head(title):
    """Returns the "head" section of HTML"""

    return """<!DOCTYPE html>
<html>
<head>
  <title>%s</title>
</head>
""" % title


def _h(title, n=1):
    """Creates "h<n>" tag"""
    return "<h%d>%s</h%d>\n" % (n, title, n)


# FITS files: <simid>+"_"+_fitss[i]+".fits"
FILE_KEYWORDS = ["cube_hr", "cube_seeing", "ifu_noseeing", "mask_fiber_in_aperture",
          "reduced", "reduced_snr", "sky", "skysub", "spintg", "spintg_noseeing",
          "therm"]


def _color(s, color):
    """Returns inline colored text"""
    return '<span style="color: %s">%s</span>' % (color, s)


def create_simulation_report(simid, dir_="."):
    """Creates HTML output and several PNG files with coherent naming

    Args:
        simid: simulation ID. This should be a string starting with a "C", e.g., "C000793"
        dir_: directory containing the simulation ouput files

    Returns:
        str: name of output file created, which will be ``<dir_>/report-<simid>.html``
    """

    if simid[0] != "C":
        raise RuntimeError("Simulation ID must start with a 'C' letter")

    # # Setup

    # file_prefix: common start to all filenames generated
    file_prefix = os.path.join(dir_, "report-"+simid)
    fn_output = file_prefix + ".html"
    # f_fn_fits = lambda middle: os.path.join(dir_, "%s_%s.fits" % (simid, middle))
    fn_log = os.path.join(dir_, "%s.out" % simid)
    fn_par = os.path.join(dir_, "%s.par" % simid)
    flag_log = os.path.isfile(fn_log)
    flag_par = os.path.isfile(fn_par)


# TODO probably load_par separate
# TODO probably load_all_fits() instead of load_bulk()
    items = load_bulk(simid, dir_)

    # My first generator!!!
    def gen_fn_fig():
        cnt = 0
        while True:
            yield "%s-%03d.png" % (file_prefix, cnt)
            cnt += 1
    next_fn_fig = gen_fn_fig()

    # # HTML Generation

    with open(fn_output, "w") as html:
        html.write(_head("Simulation %s" % simid))
        html.write("<body>")
        html.write(_h("Simulation # %s" % simid))

        html.write(_h("1. File list", 2))
        l_s = ["<pre>\n"]
        for item in items:
            l_s.append(item.filename+(" (not present)" if not item.flag_exists else "")+"\n")

        # ".par" file is an extra case
        l_s.append(fn_par)
        if not os.path.isfile(fn_par):
            l_s.append(" (not present)")
        l_s.append("\n")

        # Log file is an extra case
        l_s.append(fn_log)
        if not os.path.isfile(fn_log):
            l_s.append(" (not present)")
        l_s.append("\n")

        l_s.append("</pre>\n")
        html.write("".join(l_s))

        html.write(_h("2. Simulation specification", 2))
        if  flag_par:
            flag_par_ok = True
            try:
                filepar = ft.FilePar()
                filepar.load(fn_par)
            except Exception as E:
                flag_par_ok = False
                a99.get_python_logger().exception("Failed to load file '%s'" % fn_par)
                html.write("(" + str(E) + ")")

            if flag_par_ok:
                html.write('<table cellspacing=0 cellpadding=3 style="border: 6px solid #003000;">\n')
                for kw, va in list(filepar.params.items()):
                    html.write('<tr><td style="border-bottom: 1px solid #003000; font-weight: bold">%s</td>\n' % kw)
                    html.write('<td style="border-bottom: 1px solid #003000;">%s</td></tr>\n' % va)
                html.write("</table>\n")
        else:
            html.write(_color("File '%s' not present" % fn_par, "red"))


        # html.write(_h("2. FITS headers", 2))
        html.write(_h("3. FITS files", 2))
        html.write('<table cellspacing=0 cellpadding=3 style="border: 6px solid #003000;">\n')
        for item in items:
            html.write('<tr>\n<td colspan=2 style="vertical-align: top; border-bottom: 3px solid #003000;">\n')
            html.write("<b>%s</b>" % os.path.basename(item.filename))
            html.write('</td>\n</tr>\n')

            html.write('<tr>\n<td style="vertical-align: top; border-bottom: 6px solid #003000;">\n')
            try:
                # Opens as fits to dump header
                with fits.open(item.filename) as hdul:
                    hdul.verify()
                    try:
                        s_h = repr(hdul[0].header)
                    except:
                        # Apparently, on the second time, it works if does not on the first time
                        s_h = repr(hdul[0].header)
                html.write("<pre>%s</pre>\n" % s_h)
            except:
                a99.get_python_logger().exception("Failed to dump header from file '%s'" % item.filename)
                html.write(_color("Header dump failed", "red"))
            html.write("</td>\n")


            print(("Generating visualization for file '%s' ..." % item.filename))
            html.write('<td style="vertical-align: top; border-bottom: 6px solid #003000; text-align: center">\n')
            try:
                fig = None
                FIGURE_WIDTH = 570
                if isinstance(item.fileobj, ft.FileFullCube) and (not "cube_seeing" in item.filename):
                    # note: skips "ifu_seeing" because it takes too long to renderize
                    fig = plt.figure()
                    ax = fig.gca(projection='3d')
                    sparsecube = ft.SparseCube()
                    sparsecube.from_full_cube(item.fileobj.wcube)
                    ex.draw_cube_3d(ax, sparsecube)
                    a99.set_figure_size(fig, FIGURE_WIDTH, 480. / 640 * FIGURE_WIDTH)
                    fig.tight_layout()
                elif isinstance(item.fileobj, ft.FileSpectrum):
                    if item.keyword == "spintg":
                        sp_ref = item.fileobj.spectrum
                    fig = ex.draw_spectra([item.fileobj.spectrum])
                    a99.set_figure_size(fig, FIGURE_WIDTH, 270. / 640 * FIGURE_WIDTH)
                    fig.tight_layout()
                elif isinstance(item.fileobj, ft.FileFits):
                    if item.keyword == "mask_fiber_in_aperture":
                        fig = _draw_mask(item.fileobj)
                    elif item.keyword == "cube_seeing":
                        fig = _draw_field(item.fileobj)

                if fig:
                    fn_fig = next(next_fn_fig)
                    # print "GONNA SAVE FIGURE AS "+str(fn_fig)
                    fig.savefig(fn_fig)
                    plt.close()
                    html.write('<img src="%s"></img>' % fn_fig)
                elif item.fileobj:
                    html.write("(visualization not available for this file (class: %s)" % item.fileobj.__class__.description)
                else:
                    # TODO I already have the error information, so can improve this
                    html.write("(could not load file)")

                # html.write("hello")

            except Exception as E:
                a99.get_python_logger().exception("Failed to load file '%s'" % item.filename)
                html.write(_color("Visualization failed: "+str(E), "red"))
                html.write('<pre style="text-align: left">\n'+traceback.format_exc()+"</pre>\n")
            html.write("</td>\n")
            html.write("</tr>\n")
        html.write("</table>\n")

        html.write(_h("4. Log file dump", 2))
        if  flag_log:
            html.write("<pre>\n")
            try:
                with open(fn_log, "r") as file_log:
                    html.write(file_log.read())
            except Exception as E:
                a99.get_python_logger().exception("Failed to dump log file '%s'" % fn_log)
                html.write("("+str(E)+")")
            html.write("</pre>\n")
        else:
            html.write(_color("File '%s' not present" % fn_log, "red"))


    return fn_output


def _draw_mask(filefits):
    """Draws the image for the Cxxxxx_mask_fiber_in_aperture.fits file. Returns figure"""
    hdu = filefits.hdulist[0]
    fig = plt.figure()
    plt.imshow(hdu.data)
    plt.xlabel("pixel-x")
    plt.ylabel("pixel-y")
    HEIGHT = 300.
    nr, nc = hdu.data.shape
    width = HEIGHT/nr*nc
    a99.set_figure_size(fig, width, HEIGHT)
    plt.tight_layout()
    return fig


def _draw_field(filefits):
    """Draws data cube in grayscale. Values are calculated as 2-vector-norm"""
    hdu = filefits.hdulist[0]
    data = hdu.data
    grayscale = np.linalg.norm(data, 2, 0)
    fig = plt.figure()
    plt.imshow(grayscale, cmap='Greys_r')
    plt.xlabel("pixel-x")
    plt.ylabel("pixel-y")
    HEIGHT = 300.
    nr, nc = grayscale.shape
    width = HEIGHT/nr*nc
    a99.set_figure_size(fig, width, HEIGHT)
    plt.tight_layout()
    return fig




from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from .a_WDBMolecule import WDBMolecule
from .a_WDBState import WDBState
import os


__all__ = ["WFileMolDB"]


class WFileMolDB(a99.WBase):

    @property
    def f(self):
        return self._f

    # @f.setter
    # def f(self, x):
    #     self._f = x
    #     self.w_mol.f = x
    #     self.w_state.f = x

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._f = None
        self.flag_valid = True  # To keep XFileMainWindow happy

        # # Main layout & splitter
        lmain = self.keep_ref(QVBoxLayout(self))
        sp = self.keep_ref(QSplitter(Qt.Vertical))


        # ## Line showing the File Name
        l1 = self.keep_ref(QHBoxLayout())
        lmain.addLayout(l1)
        a99.set_margin(l1, 0)
        l1.addWidget(self.keep_ref(QLabel("<b>File:<b>")))
        w = self.label_fn = QLabel()
        l1.addWidget(w)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        # ## First widget of splitter
        w0 = self.keep_ref(QWidget())
        l0 = QVBoxLayout(w0)
        a99.set_margin(l0, 2)
        l0.setSpacing(2)

        a = self.title_mol = QLabel(a99.format_title0("Select a molecule:"))
        l0.addWidget(a)

        w = self.w_mol = WDBMolecule(self.parent_form)
        w.layout().setContentsMargins(15, 1, 1, 1)
        w.id_changed.connect(self.mol_id_changed)
        l0.addWidget(w)

        # ## Second widget of splitter
        w1 = self.keep_ref(QWidget())
        l1 = QVBoxLayout(w1)
        a99.set_margin(l1, 2)
        l1.setSpacing(2)

        a = self.title_state = self.keep_ref(QLabel(a99.format_title0("States")))
        l1.addWidget(a)

        w = self.w_state = WDBState(self.parent_form)
        w.layout().setContentsMargins(15, 1, 1, 1)
        l1.addWidget(w)

        sp.addWidget(w0)
        sp.addWidget(w1)
        lmain.addWidget(sp)

        # # Final adjustments
        a99.nerdify(self)

    def load(self, x):
        self._f = x
        self.w_mol.f = x
        self.w_state.f = x
        self.update_gui_label_fn()


    def mol_id_changed(self):
        id_ = self.w_mol.id
        row = self.w_mol.row
        self.w_state.set_id_molecule(id_)
        s = "States (no molecule selected)" if not row else "Select a State for molecule '{}'".format(
            row["formula"])
        self.title_state.setText(a99.format_title0(s))


    # # Override
    #   ========

    def update_gui_label_fn(self):
        if not self._f:
            text = "(not loaded)"
        elif self._f.filename:
            text = os.path.relpath(self._f.filename, ".")
        else:
            text = "(filename not set)"
        self.label_fn.setText(text)



__all__ = ["XToScalar"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_XHelpDialog import *
import a99
from .... import explorer as ex
import f311.filetypes as ft


_CLASS_PREFIX = ""
_CONFIG_OPERATIONS = "/gui/XToScalar/operations"


class XToScalar(XHelpDialog):
    """
    Edit Parameters to apply ToScalar blocks to a Spectrum List

    Relevant attributes:
      self.block: None or ToScalar instance, set before closing when one clicks on "OK"
      self.fieldname: string
    """

    def __init__(self, *args):
        XHelpDialog.__init__(self, *args)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        self.previous_operations = ex.get_config().get_item(_CONFIG_OPERATIONS, [])
        self.block = None

        self.setWindowTitle("Extract Scalar")

        self.labelHelpTopics.setText("&Available operations")

        # self.help_data = collect_doc(blocks.sp2scalar, base_class=blocks.baseblocks.ToScalar)
        self.help_data = a99.collect_doc(ex.blocks.toscalar, base_class=ex.ToScalar)
        self.comboBox.addItems([x[0] for x in self.help_data])
        ###
        label = QLabel(a99.enc_name_descr("O&peration", "See help below"))
        label.setAlignment(Qt.AlignRight)
        cb = self.cb_operation = QComboBox()
        label.setBuddy(cb)
        cb.setEditable(True)
        cb.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        cb.addItems(self.previous_operations)
        self.grid.addWidget(label, 0, 0)
        self.grid.addWidget(cb, 0, 1)
        # ###
        # edit = self.editFunction = QLineEdit("ToScalar_SNR(0, 100000)")
        # label.setBuddy(edit)
        # self.grid.addWidget(label, 0, 0)
        # self.grid.addWidget(edit, 0, 1)
        label = keep_ref(QLabel(a99.enc_name_descr("&Target Field Name", "Leave blank to use Function Name")))
        label.setAlignment(Qt.AlignRight)
        edit = self.editFieldName = QLineEdit("")
        label.setBuddy(edit)
        self.grid.addWidget(label, 1, 0)
        self.grid.addWidget(edit, 1, 1)

        a99.place_center(self, 800, 600)


    def accept(self):
        try:
            expr = str(self.cb_operation.currentText())
            symbols_available = a99.module_to_dict(ex.blocks.toscalar)
            import numpy as np
            symbols_available["np"] = np
            block = eval(expr.strip(), symbols_available)

            if not isinstance(block, ex.ToScalar):
                raise RuntimeError("Expression does not evaluate to a valid ToScalar Block")

            self.block = block
            fieldname = str(self.editFieldName.text()).strip()
            fieldname = a99.expr_to_fieldname(expr) if len(fieldname) == 0 else fieldname
            self.fieldname = a99.valid_fits_key(fieldname)

            # Saves new operation on close (if evaluated successfully)
            if not expr in self.previous_operations:
                self.previous_operations.append(expr)
                self.previous_operations.sort()
                ex.get_config().set_item(_CONFIG_OPERATIONS, self.previous_operations)

            return QDialog.accept(self)
        except Exception as e:
            self.add_log_error(a99.str_exc(e), True)
            return False



__all__ = ["SparseCube"]


import a99
from . import SpectrumCollection, FullCube
import numpy as np
from scipy.interpolate import interp1d
from astropy.io import fits


# headers to care about when importing from a FullCube HDU
_HEADERS_COMPASS_CUBE = ["CDELT1", "HRFACTOR", "R"]


@a99.froze_it
class SparseCube(SpectrumCollection):
    """
    Spectral cube allocated in a way that will take less space.

    Each spectrum has its (x, y, z) coordinate for placement inside the cube
    and is layed from this point along the z-axis (i.e., wavelength axis)

    Saved in FITS format, as it is small and can be opened by other programs
    """

    attrs = SpectrumCollection.attrs + ["R", "hrfactor", "hr_pix_size"]

    @property
    def hr_pix_size(self):
        return self.more_headers.get("CDELT1")

    @hr_pix_size.setter
    def hr_pix_size(self, value):
        self.more_headers["CDELT1"] = value

    @property
    def hrfactor(self):
        return self.more_headers.get("HRFACTOR")

    @hrfactor.setter
    def hrfactor(self, value):
        self.more_headers["HRFACTOR"] = value

    @property
    def R(self):
        return self.more_headers.get("R")

    @R.setter
    def R(self, value):
        self.more_headers["R"] = value

    @property
    def width(self):
        return self.more_headers.get("WIDTH")

    @width.setter
    def width(self, value):
        self.more_headers["WIDTH"] = value

    @property
    def height(self):
        return self.more_headers.get("HEIGHT")

    @height.setter
    def height(self, value):
        self.more_headers["HEIGHT"] = value

    # read-only
    @property
    def delta_lambda(self):
        return self.wavelength[1] - self.wavelength[0]

    def __init__(self, hdu=None):
        SpectrumCollection.__init__(self)
        self.__flag_update = True
        self.__flag_update_pending = False
        self.wavelength = np.array(
            [-1., -1.])  # Like this delta_lambda property does not give error
        # Header data initialized to default
        self.R = 5000  # TODO this does not match the spectrum R necessarily, should I do sth about it?
        self.hrfactor = 10
        self.hr_pix_size = 0.0375 / self.hrfactor
        self.width = 50
        self.height = 30

        if hdu is not None:
            self.from_hdu(hdu)

    def crop(self, x0=None, x1=None, y0=None, y1=None, lambda0=None, lambda1=None):
        """
        Reduces region, discards spectra, cuts remaining spectra

        intervals:
            X - [x0, x1]
            Y - [y0, y1]
            wavelength - [lambda0, lambda1]

        **Note** x1, y1, lambda 1 **included** in interval (not pythonic).
        """
        if len(self.spectra) == 0:
            raise RuntimeError("Need at least one spectrum added in order to crop")

        if x0 is None:
            x0 = 0
        if x1 is None:
            x1 = self.width - 1
        if y0 is None:
            y0 = 0
        if y1 is None:
            y1 = self.height - 1
        if lambda0 is None:
            lambda0 = self.wavelength[0]
        if lambda1 is None:
            lambda1 = self.wavelength[-1]

        if x0 > x1:
            raise RuntimeError('x0 is greater than x1')
        if y0 > y1:
            raise RuntimeError('y0 is greater than y1')
        if lambda0 > lambda1:
            raise RuntimeError('lambda0 is greater than lambda1')

        if not any([x0 != 0, x1 != self.width - 1, y0 != 0, y1 != self.height - 1,
                    lambda0 != self.wavelength[0], lambda1 != self.wavelength[-1]]):
            return

        if any([x0 != 0, x1 != self.width - 1, y0 != 0, y1 != self.height - 1]):
            self.width = x1 - x0 + 1
            self.height = y1 - y0 + 1
            for sp in self.spectra:
                sp.x -= x0
                sp.y -= y0

            # Deletes spectra out of XY range
            for sp in reversed(self.spectra):
                if not 0 <= sp.pixel_x < self.width or not 0 <= sp.pixel_y < self.height:
                    self.spectra.remove(sp)

        if any([lambda0 != self.wavelength[0], lambda1 != self.wavelength[-1]]):
            # cuts remaining spectra
            for sp in reversed(self.spectra):
                sp.cut(lambda0, lambda1)
                if len(sp) == 0:
                    # If spectrum is now out of range, it is deleted
                    self.spectra.remove(sp)
        self.__update()

    def from_hdulist(self, hdul):
        """
        Checks for required field names existence

        Toad successfully, file must have the following field names:
        'PIXEL-X', 'PIXEL-Y', 'Z-START'
        """

        self.__flag_update = False
        try:
            # if not hdul[0].header.get("TAINHA"):
            #     raise RuntimeError("Wrong HDUList")

            SpectrumCollection.from_hdulist(self, hdul)

            required = ('PIXEL-X', 'PIXEL-Y', 'Z-START')
            for fn in required:
                if not fn in self.fieldnames:
                    raise RuntimeError("Required field name '%s' not found" % fn)

                    #            for i, hdu in enumerate(hdul):
                    #                if i == 0:
                    #                    for na0, na1 in _MMM:
                    #                        try:
                    #                            self.__setattr__(na0, hdu.header[na1])
                    #                        except:
                    #                            a99.get_python_logger().exception("Failed setting '%s' = '%s'" % (na0, na1))
                    #
                    #                else:
                    #                    sp = Spectrum()
                    #                    sp.from_hdu(hdu)
                    #                    if i == 1:
                    #                        # TODO this must be settable, not just taken from first spectrum
                    #                        self.reference = _LambdaReference(sp.x)
                    #                    self.add_spectrum(sp)
        finally:
            self.enable_update()

    # def to_hdulist(self):
    #     """Inherited to add FileSparseCube-specific header marker"""
    #     hdul = SpectrumCollection.to_hdulist(self)
    #     hdul[0].header["TAINHA"] = 26.9752
    #     return hdul
    #
    def from_full_cube(self, full_cube):
        assert isinstance(full_cube, FullCube)
        hdu = full_cube.hdu
        assert isinstance(hdu, fits.PrimaryHDU)
        data = hdu.data
        nlambda, nY, nX = data.shape

        # Reads some attributes from the headers
        # Uses self.attrs, but this is just a coincidence, may need detachment
        for name in _HEADERS_COMPASS_CUBE:
            self.more_headers[name] = hdu.header.get(name, self.more_headers[name])

        self.__flag_update = False
        try:
            # TODO implement threshold, i.e., abs(...) <= epsilon
            xx, yy = np.where(sum(data, 0) != 0)

            for i, j in zip(xx, yy):
                sp = full_cube.get_spectrum(i, j)
                # discards edges that are zeros
                where_positive = np.where(sp.flux > 0)[0]
                sp.cut_idxs(where_positive[0], where_positive[-1] + 1)
                sp.pixel_x, sp.pixel_y = i, j
                self.add_spectrum(sp)

            self.height = nY
            self.width = nX
        finally:
            self.enable_update()

    def to_full_cube(self):
        """Creates FullCube object"""
        assert len(self.spectra) > 0, "No spectra added"

        wcube = FullCube()
        wl_new = wcube.wavelength = self.wavelength.copy()
        dims = len(wl_new), self.height, self.width
        wcube.create1(self.R, dims, self.hr_pix_size, self.hrfactor)
        for sp in self.spectra:
            ii0 = a99.BSearchCeil(wl_new, sp.x[0])
            wcube.hdu.data[ii0:ii0 + len(sp), sp.pixel_y, sp.pixel_x] = sp.y
        wcube.set_wavelength(self.wavelength)
        return wcube


    def add_spectrum(self, sp):
        """
        "Paints" pixel with given spectrum

        Args:
            sp: Spectrum instance

        **Note** coordinate (x=0, y=0) corresponds to lower left pixel of cube cross-section
        """
        assert isinstance(sp, ft.Spectrum)
        # assert self.flag_created, "Cube has not been created yet"

        if len(sp.x) < 2:
            raise RuntimeError("Spectrum must have at least two points")

        sp.start_z = -1  # will cause update to check on spectrum
        SpectrumCollection.add_spectrum(self, sp)
        self.__update()

    #    def delete_spectra(self, indexes):
    #        indexes = list(set(indexes))
    #        if isinstance(indexes, numbers.Integral):
    #            indexes = [indexes]
    #        for index in reversed(indexes):
    #            del self.spectra[index]
    #        self.__update()

    def enable_update(self):
        self.__flag_update = True
        if self.__flag_update_pending:
            self.__update()
            self.__flag_update_pending = False

    def get_pixel(self, x, y, flag_copy=True):
        """Creates spectrum on-the-fly containing all spectra merged for pixel (x, y)

        **Note** if there is no spectra at point (x, y), returns an empty spectrum, for speed

        Args:
            x, y: 0-based pixel coordinates
            flag_copy: disable vector copies to speed up but don't the spectrum
        """

        ret = ft.Spectrum()
        if len(self.spectra) > 0:
            any_ = False
            for sp in self.spectra:
                if sp.pixel_x == x and sp.pixel_y == y:
                    if not any_:
                        ret.x = self.wavelength if not flag_copy else np.copy(self.wavelength)
                        ret.y = np.zeros(len(ret.x))
                        any_ = True
                    ii0 = a99.BSearchCeil(ret.x, sp.x[0])
                    ret.y[ii0:ii0 + len(sp)] = sp.y
        return ret

    def __update(self):
        """Updates internal state

        - resamples necessary vectors
        - spectra z-positions
        - wavelength vector
        """

        if not self.__flag_update:
            self.__flag_update_pending = True
            return

        self.wavelength = np.array([-1., -1.])

        if len(self.spectra) == 0:
            return

        wlmax = max([sp.x[-1] for sp in self.spectra])
        wlmin = min([sp.x[0] for sp in self.spectra])
        sp = self.spectra[0]
        dl = sp.x[1] - sp.x[0]
        wlmin = (wlmin // dl) * dl
        wlmax = np.ceil(wlmax / dl) * dl
        wl = self.wavelength = np.arange(wlmin, wlmax + dl, dl)

        for sp in self.spectra:
            i0 = a99.BSearchCeil(wl, sp.x[0])
            assert i0 != -1, "BSearchCeil(wl, sp.x[0]) FAILED"
            i1 = a99.BSearchFloor(wl, sp.x[-1])
            assert i1 != -1, "BSearchFloor(wl, sp.x[-1]) FAILED"

            # TODO account for different delta lambda
            if sp.delta_lambda != dl or wl[i0] != sp.x[0] or wl[i1] != sp.x[-1]:
                # # Linear interpolation: either adjust for different delta lambda or misaligned wavelengths
                wl_new = wl[i0:i1 + 1]
                f = interp1d(sp.x, sp.y)
                sp.x = wl_new
                sp.y = f(wl_new)
            sp.z_start = i0




"""VALD3 atomic or molecular lines file"""


__all__ = ["KuruczMolLine", "FileKuruczMolecule"]

# from ..gear import *
import sys
import a99
from .. import DataFile
import io
import fortranformat as ff
import os


@a99.froze_it
class KuruczMolLine(a99.AttrsPart):
    attrs = ["lambda_", "loggf", "J2l", "E2l", "Jl", "El", "atomn0", "atomn1", "state2l", "v2l",
             "lambda_doubling2l", "spin2l", "statel", "vl", "lambda_doublingl", "spinl", "iso",]

    def __init__(self, lambda_=None, loggf=None, J2l=None, E2l=None, Jl=None, El=None, atomn0=None,
                 atomn1=None, state2l=None, v2l=None, lambda_doubling2l=None, spin2l=None,
                 statel=None, vl=None, lambda_doublingl=None, spinl=None, iso=None):
        a99.AttrsPart.__init__(self)
        # Wavelength in **Angstrom*** (although it is is stored in nm)
        self.lambda_ = lambda_
        self.loggf = loggf
        self.J2l = J2l
        self.E2l = E2l
        self.Jl = Jl
        self.El = El
        self.atomn0 = atomn0
        self.atomn1 = atomn1
        self.state2l = state2l
        self.v2l = v2l
        self.lambda_doubling2l = lambda_doubling2l
        self.spin2l = spin2l
        self.statel = statel
        self.vl = vl
        self.lambda_doublingl = lambda_doublingl
        self.spinl = spinl
        self.iso = iso

    def __repr__(self):
        return "{}({}, {}, {}, {}, {}, {}, {}, {}, '{}', {}, '{}', {}, '{}', {}, '{}', {}, {})".\
            format(self.__class__.__name__, self.lambda_,
            self.loggf, self.J2l, self.E2l, self.Jl, self.El, self.atomn0, self.atomn1,
            self.state2l, self.v2l, self.lambda_doubling2l, self.spin2l, self.statel, self.vl,
            self.lambda_doublingl, self.spinl, self.iso,)


class FileKuruczMolecule(DataFile):
    """
    Kurucz molecular lines file

    **Note** Kurucz file so far refers to one molecule only. Apparently one single system as well,
             although we will record the system information (A, X etc)

    **Note** Load only
    """

    attrs = ["num_lines"]

    @property
    def num_lines(self):
        return len(self)

    def __len__(self):
        return len(self.lines)

    def __init__(self):
        DataFile.__init__(self)

        # list of KuruczMolLine objects
        self.lines = []

    def __iter__(self):
        return iter(self.lines)


    def _do_load(self, filename):

        # **note** Kurucz puts always the "double-line" values before the "line" values
        #
        #   Wl(nm)   loggf   J"    E(cm-1)  J'   E(cm-1)   H
        #   |        |       |     |        |    |         |O
        #   |        |       |     |        |    |         || electronic state
        #   |        |       |     |        |    |         || |v"=00
        #   |        |       |     |        |    |         || || lambda-doubling component
        #   |        |       |     |        |    |         || || |spin
        #   |        |       |     |        |    |         || || ||
        #   204.5126 -7.917  2.5    83.925  2.5  48964.990 108X00f1   A07e1   16
        #   204.7561 -7.745  3.5   202.380  3.5  49025.320 108X00f1   A07e1   16
        #   204.9400 -7.883  5.5   543.596  6.5  49322.740 108X00e1   A07e1   16
        #   205.0076 -7.931  3.5   201.931  2.5  48964.990 108X00e1   A07e1   16
        #   205.0652 -7.621  4.5   355.915  4.5  49105.280 108X00f1   A07e1   16
        #   205.0652 -7.621  4.5   355.915  4.5  49105.280 108X00f1   A07e1   16
        # 1         11     18   23        33   38         49          61      69 1-based
        # 0         10     17   22        32   37         48          60      68 0-based (pythonic)
        #
        # FORMAT(F10.4.F7.3,F5.1,F10.3,F5.1,F11.3,I4,A1,I2,A1,I1,3X,A1,I2,A1,I1,3X,I2)
        # "
        # The code for the diatomic molecules is two 2-digit element numbers in
        # ascending order.  The labels consist of the electronic state, the vibrational
        # level, the lambda-doubling component, and the spin state.  Sometimes two
        # characters are required for the electronic state and the format becomes
        # ,A2,I2,A1,i1,2X,.  Negative energies are predicted or extrapolated"
        #
        # http://kurucz.harvard.edu/linelists/0linelists.readme

        filesize = os.path.getsize(filename)
        num_lines = int(filesize/70)

        with open(filename, "r") as h:
            self._do_load_h(h, filename, num_lines)

    def _do_load_h(self, h, filename, num_lines=0):
        # Thank you Kurucz
        fr = ff.FortranRecordReader(
            '(F10.4,F7.3,F5.1,F10.3,F5.1,F11.3,I2,I2,A1,I2,A1,I1,3X,A1,I2,A1,I1,3X,I2)')
        r = 0  # counts rows of file
        ii = 0
        try:
            self.lines = []
            while True:
                s = h.readline().strip("\n")
                if len(s) == 0:
                    break

                line = KuruczMolLine()
                # FortranFormat too slow
                # line.lambda_, line.loggf, line.J2l, line.E2l, line.Jl, line.El, line.atomn0, \
                # line.atomn1, line.state2l, line.v2l, line.lambda_doubling2l, line.spin2l, \
                # line.statel, line.vl, line.lambda_doublingl, line.spinl, line.iso = fr.read(s)

                #
                line.lambda_ = float(s[0:10])*10
                line.loggf = float(s[10:17])
                line.J2l = float(s[17:22])
                line.E2l = float(s[22:32])
                line.Jl = float(s[32:37])
                line.El = float(s[37:48])
                line.atomn0 = int(s[48:50])
                line.atomn1 = int(s[50:52])
                line.state2l = s[52:53]
                line.v2l = int(s[53:55])
                line.lambda_doubling2l = s[55:56]
                line.spin2l = int(s[56:57])
                line.statel = s[60:61]
                line.vl = int(s[61:63])
                line.lambda_doublingl = s[63:64]
                line.spinl = int(s[64:65])
                line.iso = s[69:71]
                self.lines.append(line)
                r += 1
                ii += 1
                if ii == 103:
                    a99.get_python_logger().info(
                        "Loading '{}': {}".format(filename, a99.format_progress(r, num_lines)))
                    ii = 0


        except Exception as e:
            # f = type(e)(("Error around %d%s row of file '%s'" %
            #              (r + 1, a99.ordinal_suffix(r + 1), filename)) + ": " + str(
            #     e)).with_traceback(sys.exc_info()[2])
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename, a99.str_exc(e))) from e


def _fake_file():
    """Returns StringIO for a fragment of a Kurucz molecular lines file"""

    f = io.StringIO()
    f.write("""  204.5126 -7.917  2.5    83.925  2.5  48964.990 108X00f1   A07e1   16
  204.7561 -7.745  3.5   202.380  3.5  49025.320 108X00f1   A07e1   16
  204.9400 -7.883  5.5   543.596  6.5  49322.740 108X00e1   A07e1   16
  205.0076 -7.931  3.5   201.931  2.5  48964.990 108X00e1   A07e1   16
  205.0652 -7.621  4.5   355.915  4.5  49105.280 108X00f1   A07e1   16
""")

    f.seek(0)
    return f


def _test():
    """Test code in docstring

    Example:

    >>> h = _fake_file()
    >>> f = FileKuruczMolecule()
    >>> f._do_load_h(h, "_fake_file")
    >>> f.lines[0]
    KuruczMolLine(2045.126, -7.917, 2.5, 83.925, 2.5, 48964.99, 1, 8, 'X', 0, 'f', 1, 'A', 7, 'e', 1, 6)
    """
    return

# h = _fake_file()
# f = FileVald3()
# f._do_load_h(h, "_fake_file")



class MolConversionLog(object):
    def __init__(self, num_lines_in=0, errors=[], flag_ok=True):
        self.errors = errors
        # Number of input lines
        self.num_lines_in = num_lines_in
        # Whether the conversion was considered successful (despite errors)
        self.flag_ok = flag_ok

    def __repr__(self):
        return "{}({}, {}, {})".format(self.__class__.__name__, self.num_lines_in, self.errors,
                                       self.flag_ok)
    


import sys
import os

if sys.version_info[0] < 3:
    print("Python version detected:\n*****\n{0!s}\n*****\nCannot run, must be using Python 3".format(sys.version))
    sys.exit()

from setuptools import setup, find_packages
from glob import glob


def find_scripts(pkgnames):
    ret = []
    for pkgname in pkgnames:
        for item in os.walk(pkgname):
            _, last = os.path.split(item[0])
            if last == "scripts":
                ret.extend(glob(os.path.join(item[0], '*.py')))
    return ret


pkgs = find_packages()
scripts = find_scripts(["f311"])


setup(
    name='f311',
    packages=find_packages(),
    include_package_data=True,
    version='0.17.2.16',
    license='GNU GPLv3',
    platforms='any',
    description='Collection of Astronomy-related libraries and applications in Python',
    author='Julio Trevisan',
    author_email='juliotrevisan@gmail.com',
    url='https://github.com/trevisanj/f311', # use the URL to the github repo
    keywords= ['astronomy', "fits", "spectroscopy", "spectral synthesis", "photometry",
               "honl-london", "nist", "hitran", "multiplicity"],
    install_requires=[],
    scripts=scripts
)


# TODO later install_requires=['numpy', 'matplotlib', 'pyqt5'],  # matplotlib never gets installed correctly by pip, but anyway...



from .a_XHelpDialog import *
from .a_XPlotXY import *
from .a_XPlotXYZ import *
from .a_WFileSparseCube import *
from .a_WFileSpectrumList import *
from .a_WSpectrumCollection import *
from .a_XFileSparseCube import *
from .a_XFileSpectrumList import *
from .a_XGroupSpectra import *
from .a_XScaleSpectrum import *
from .a_XToScalar import *
from .a_XUseSpectrumBlock import *
# from . import a_WFileSparseCube
# from . import a_WFileSpectrumList
# from . import a_WSpectrumCollection
# from . import a_XFileSparseCube
# from . import a_XFileSpectrumList
# from . import a_XGroupSpectra
# from . import a_XHelpDialog
# from . import a_XPlotXY
# from . import a_XPlotXYZ
# from . import a_XScaleSpectrum
# from . import a_XToScalar
# from . import a_XUseSpectrumBlock


""""
List of spectra sharing same wavenumber axis. Uses FITS format
"""


__all__ = ["SpectrumList"]


from . import SpectrumCollection, FullCube
# from a99 import froze_it, cut_spectrum
from .. import Spectrum
from astropy.io import fits
import numpy as np
import a99


@a99.froze_it
class SpectrumList(SpectrumCollection):
    attrs = SpectrumCollection.attrs+["wavelength"]

    @property
    def delta_lambda(self):
        return self.wavelength[1]-self.wavelength[0]

    @property
    def flag_wled(self):
        """Wavelength problem already resolved?"""
        return self.wavelength[0] > -1

    def __init__(self, hdulist=None):
        SpectrumCollection.__init__(self)
        self.__flag_update = True
        self.__flag_update_pending = False
        self.wavelength = np.array([-1., -1.])  # the wavelength axis (angstrom) (shared among all spectra in the cube)

        if hdulist is not None:
            self.from_hdulist(hdulist)

    ############################################################################
    # # Interface

    def to_csv(self, sep=","):
        """Generates tabulated text

        Returns list of strings
        """
        lines = []
        lines.append(sep.join(self.fieldnames+[str(x)  for x in self.wavelength])+"\n")
        for sp in self.spectra:
            lines.append(sep.join([repr(sp.more_headers.get(fieldname)) for fieldname in self.fieldnames]+
                                  [str(flux) for flux in sp.y])+"\n")
        return lines


    def matrix(self):
        """Returns a (spectrum)x(wavelength) matrix of flux values"""
        n = len(self)
        if n == 0:
            return np.array()
        return np.vstack([sp.y for sp in self.spectra])

    def from_hdulist(self, hdul):
        self.__flag_update = False
        try:
            SpectrumCollection.from_hdulist(self, hdul)
        finally:
            self.enable_update()

    def from_full_cube(self, full_cube):
        """
        Adds all cube "pixels" (i.e., spectra) that are not all zero

        Very similar to SparseCube.from_full_cube() (bit simpler)
        """

        assert isinstance(full_cube, FullCube)
        hdu = full_cube.hdu
        assert isinstance(hdu, fits.PrimaryHDU)
        data = hdu.data

        # TODO implement threshold, i.e., abs(...) <= epsilon
        xx, yy = np.where(sum(data, 0) != 0)
        for i, j in zip(xx, yy):
            sp = full_cube.get_spectrum(i, j)
            sp.pixel_x, sp.pixel_y = i, j
            # TODO have to fill in filename first sp.more_headers["ORIGIN"] = os.path.basename(self.filename)
            self.add_spectrum(sp)


    def add_spectrum(self, sp):
        """Adds spectrum. Updates internal wavelength vector to maximum possible

        If wavelength vectors do not match, it will resample the new spectrum,
        and may expand self.wavelength, but will not shift the x-position of existing
        points
        """
        assert isinstance(sp, Spectrum)
        if len(sp.x) < 2:
            raise RuntimeError("Spectrum must have at least two points")

        if not self.flag_wled:
            self.wavelength = np.copy(sp.wavelength)
        else:
            if not np.all(self.wavelength == sp.wavelength):
                # print("VAI TER QUE RESAMPLEAR ALGO")
                xcur0, xcur1 = self.wavelength[0], self.wavelength[-1]
                xsp0, xsp1 = sp.x[0], sp.x[-1]

                # quantizes new wavelength interval to current delta_lambda step
                dl = self.delta_lambda

                xnew0 = xcur0 if xcur0 <= xsp0 else xcur0-np.floor((xcur0-xsp0)*dl)/dl
                xnew1 = xcur1 if xcur1 >= xsp1 else xcur1+np.floor((xsp1-xcur1)*dl)/dl

                n = int(np.round((xnew1-xnew0)/dl))+1
                self.wavelength = np.arange(n)*dl+xnew0

                if not (xnew0 == xcur0 and xnew1 == xcur1):
                    # print("RESAMPLEANDO EXISTING")
                    for sp_existing in self.spectra:
                        sp_existing.resample(self.wavelength)

                if not(xnew0 == xsp0 and xnew1 == xsp1 and dl == sp.delta_lambda):
                    # print("RESAMPLING NEWCOMER")
                    sp.resample(self.wavelength)
                else:
                    # print("NO NEED TO RESAMPLE NEWCOMER")
                    pass

                # raise RuntimeError("Cannot add spectrum, wavelength vector does not match existing")

        SpectrumCollection.add_spectrum(self, sp)
        self.__update()

    def delete_spectra(self, indexes):
        SpectrumCollection.delete_spectra(self, indexes)
        self.__update()

    def enable_update(self):
        self.__flag_update = True
        if self.__flag_update_pending:
            self.__update()
            self.__flag_update_pending = False


    def __update(self):
        """
        Updates internal state

        This consists of verifying whether or not there are spectra.
        If not, resets the wavelength vector.
        """

        if not self.__flag_update:
            self.__flag_update_pending = True
            return

        if len(self.spectra) == 0:
            self.wavelength = np.array([-1., -1.])



"""
Miscellanea of basic routines used to implement the file classes
"""

import re
import os
import shutil
import sys


__all__ = ["adjust_atomic_symbol", "description_to_symbols", "iz_to_branch", "branch_to_iz",
           "get_default_data_path"]


def adjust_atomic_symbol(x):
    """Makes sure atomic symbol is right-aligned and upper case (PFANT convention)."""
    assert isinstance(x, str)
    return "%2s" % (x.strip().upper())

# Formulas that can be recognized in PFANT molecule header information ("titulo")
# This exists for backward compatibility. New molecules should have its symbols informed separately
#
# **Careful** If you add entried, make sure everything is in uppercase
_BUILTIN_FORMULAS = {
     "MGH": ["MG", "H"],
     "C2": ["C", "C"],
     "CN": ["C", "N"],
     "CH": ["C", "H"],
     "13CH": ["A", "H"],
     "12C16O": ["C", "O"],
     "NH": ["N", "H"],
     "OH": ["O", "H"],
     "FEH": ["FE", "H"],
     "TIO": ["TI", "O"],
     "CO": ["C", "O"],
}


def description_to_symbols(descr):
    """Finds molecular formula in description text and returns separate symbols

    This routine is similar to pfantlib.f90:assign_symbols_by_formula()

    It is used when the symbols are not informed separately
    after the first "#" in the molecule 'titulo'

    Args:
        descr: string

    Returns:
        list of two elements, e.g., [" C", " O"]

    Example:

    >>> import f311.filetypes as ft
    >>> ft.description_to_symbols("12C16O INFRARED  X 1 SIGMA+,   version 15/oct/98")
    [' C', ' O']
    """

    descr = descr.upper()
    for formula, symbols in _BUILTIN_FORMULAS.items():
        # Searches for formula at start or after whitespace
        if re.search(r'(^|\W){}([^0-9A-Z]|$)'.format(formula), descr):
            return [adjust_atomic_symbol(x) for x in symbols]
    return None
    # raise RuntimeError("Could not find a valid formula in description '{}'".format(descr))


_iz_to_branch_map = {"1": "P", "2": "Q", "3": "R", "4": "P1", "5": "Q1", "6": "R1", "7": "P2",
                     "8": "Q2", "9": "R2", "10": "P3", "11": "Q3", "12": "R3", }
_branch_to_iz_map = dict(((value, key) for key, value in _iz_to_branch_map.items()))


def iz_to_branch(iz):
    """Converts BLB's 'iz' code to string P/Q/R/P1 ... (see pfantlib.90:read_molecules()"""
    return _iz_to_branch_map[str(iz)]


def branch_to_iz(br):
    """Converts branch P/Q/R/P1, etc. into BLB's 'iz' code"""
    return _branch_to_iz_map[br]


def get_default_data_path(*args, module=None, class_=None):
    """
    Returns path to default data directory

    Arguments 'module' and 'class' give the chance to return path relative to package other than
    f311.filetypes

    Args:
        module: Python module object. It is expected that this module has a sub-subdirectory
                named 'data/default'
        class_: Python class object to extract path information from. If this argument is used,
                it will be expected that the class "root" package will have a subpackage called
                'filetypes', i.e., '(some package).filetypes'.
                **Has precedence over 'module' argument**

    """
    if module is None:
        module = __get_filetypes_module()

    if class_ is not None:
        pkgname =  class_.__module__
        mseq = pkgname.split(".")
        if len(mseq) < 2 or mseq[1] != "filetypes":
            raise ValueError("Invalid module name for class '{}': '{}' "
                             "(must be '(...).filetypes[.(...)]')".format(class_.__name__, pkgname))
        module = sys.modules[".".join(mseq[:2])]
    module_path = os.path.split(module.__file__)[0]
    p = os.path.abspath(os.path.join(module_path, "data", "default", *args))
    return p


def copy_default_data_file(filename, module=None):
    """Copies file from ftpyfant/data/default directory to local directory."""
    if module is None:
        module = __get_filetypes_module()
    fullpath = get_default_data_path(filename, module=module)
    shutil.copy(fullpath, ".")


def __get_filetypes_module():
    from f311 import filetypes as ft
    return ft




"""
Class Catalog -- resources to retrieve File* classes by different criteria
"""

from collections import OrderedDict
# from .. import filetypes as ft
import importlib
import a99

__all__ = [
    "classes_txt", "classes_bin", "classes_sp", "classes_file", "classes_collection"
    ]


__COLLABORATORS = ["f311.filetypes"]


# def collaborators():
#     """Returns a dictionary of packages scanned to populate _classes_*
#
#     Example: {"f311.filetypes": f311.filetypes, }
#     """
#     if __flag_first:
#         __setup()
#     return __collaborators


def classes_txt():
    """Classes to consider when attempts to load a text file (see load_any_file())"""
    if __flag_first:
        __setup()
    return _classes_txt


def classes_bin():
    """Classes to consider when attempts to load a binary file (see load_any_file())"""
    if __flag_first:
        __setup()
    return _classes_bin


def classes_sp():
    """Classes to consider when attempts to load a spectrum file (see load_spectrum())"""
    if __flag_first:
        __setup()
    return _classes_sp


def classes_file():
    """All known File* classes"""
    if __flag_first:
        __setup()
    return _classes_file


def classes_collection():
    """
    Returns list of File* classes that can be converted to a SpectrumCollection
    """
    from f311 import filetypes as ft
    return classes_sp() + [ft.FileSpectrumList, ft.FileSparseCube, ft.FileFullCube]


def _collect_classes(m):
    """
    Adds entries to _classes_*

    Args:
        m: module object that must contain the following sub-modules: datatypes, vis
    """
    from f311 import filetypes as ft

    def _extend(classes, newclasses):
        """Filters out classes already present in list.

        This shouldn't be necessary, but collaborators may accidentally import already loaded
        classes into the datatypes namespace"""
        classes.extend([class_ for class_ in newclasses if class_ not in classes])
        # classes.extend(newclasses)

    file_classes = a99.get_classes_in_module(m, ft.DataFile)

    # Classes to consider when attempts to load a text file (see load_any_file())
    _extend(_classes_txt, [class_ for class_ in file_classes if class_.flag_txt])

    # Classes to consider when attempts to load a binary file (see load_any_file())
    _extend(_classes_bin, [class_ for class_ in file_classes if not class_.flag_txt])
    # Adds Classes to consider when attempts to load a spectrum file (see load_spectrum())
    _extend(_classes_sp, [class_ for class_ in file_classes if issubclass(class_, ft.FileSpectrum)])
    # All kwown File* classes
    _extend(_classes_file, file_classes)


# # List of classes representing all file formats either read or written
#   ====================================================================
_classes_txt = []
_classes_bin = []
_classes_sp = []
_classes_file = []
__flag_first = True
__collaborators = OrderedDict()  # (("f311.filetypes",  f311.filetypes),))


def __setup():
    """Will be executed in the first time someone calls classes_*() """
    global __collaborators, __flag_first
    __flag_first = False
    for pkgname in __COLLABORATORS:
        try:
            pkg = importlib.import_module(pkgname)
            a99.get_python_logger().info("Imported collaborator package '{}'".format(pkgname))

            try:
                if hasattr(pkg, "_setup_filetypes"):
                    pkg._setup_filetypes()
                else:
                    _collect_classes(pkg)

                __collaborators[pkgname] = pkg
            except:
                a99.get_python_logger().exception(
                    "Actually, package '{}' gave error".format(pkgname))
                raise
        except:
            a99.get_python_logger().warning("Failed to import package '{}".format(pkgname))
            pass



#!/usr/bin/env python3


"""
Downloads WebSim-COMPASS simulations

Based on shell script by Mathieu Puech

**Note** Skips simulations for existing files in local directory starting with
         that simulation ID.
         Example: if it finds file(s) "C001006*", will skip simulation C001006

**Note** Does not create any directory (actually creates it but deletes later).
         All files stored in local directory!

**Note** Will work only on if os.name == "posix" (Linux, UNIX ...)
"""


import os
import argparse
import sys
import glob
import logging
from f311 import aosss
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True



def print2(*args):
  a99.get_python_logger().info(*args)


if __name__ == "__main__":
    if os.name != "posix":
        print2("OS is '"+os.name+"', this script is only for 'posix' OS's, sorry.")
        sys.exit()

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument('--max', metavar='N', type=int, default=100,
                        help='Maximum number of simulations to get')
    parser.add_argument('--stage', type=str, nargs='?', default="all",
     help="Websim-Compass pipeline stage: if specified, will download files named, e.g., "
          "C000793_<stage>.fits (**note**: .par and .out files are always downloaded)")

    parser.add_argument('numbers', metavar='N', type=str, nargs='+',
                     help='List of simulation numbers (single value and ranges accepted, e.g. 1004, 1004-1040)')
    args = parser.parse_args()

    simids = aosss.compile_simids(args.numbers)
    print2("List of simulation IDs: "+", ".join(simids))

    if len(simids) > args.max:
        print2("Number of simulations to download (%d) exceeds maximum number (%d)" % (len(simids), args.max))
        print2("Use --max option to raise this maximum number")
        sys.exit()

    # Part based on Mathieu Puech's script
    #
    # wget options explained:
    #   -v       -- verbose
    #   -P       --  prefix: directory to save files to
    #   -A       -- accept: accept only certain file types
    #   -r       -- recursive
    #   -l1      -- specify maximum recursion level depth
    #   -e       -- execute a command as if it were part of .wgetrc
    #   -np      -- (--no-parent)
    #               do not ever ascend to the parent directory when retrieving recursively
    site = "http://websim-compass.obspm.fr/wcompass"
    for simid in simids:
        ff = glob.glob("%s*" % simid)
        if len(ff) > 0:
            print2("Files found starting with '%s', skipping this download" % simid)
        print2("Downloading %s simulation files..." % simid)
        # os.system("wget -v -r -l1 -np -nd -erobots=off -P %s %s/%s/" % (simid, site, simid))
        os.system("wget -v -P %s %s/%s.out" % (simid, site, simid))
        os.system("wget -v -P %s %s/%s.par" % (simid, site, simid))

        if args.stage == "all":
            os.system("wget -v -r -l1 -np -nd -erobots=off -P %s -A.fits -A.par %s/%s/" % (simid, site, simid))
        else:
            os.system("wget -v -P {0} {1}/{0}/{0}_{2}.fits".format(simid, site, args.stage))

        os.system("mv %s/* ." % simid)
        os.system("rm -r %s" % simid)




__all__ = ["GB_UseNumPyFunc", "GB_SNR"]


from .basic import *
from . import sb
import a99
from ... import explorer as ex

class GB_UseNumPyFunc(GroupBlock):
    """Output contains single spectrum whose y-vector is calculated using a numpy function

    The numpy function must be able to operate on the first axis, e.g., np.mean(), np.std()
    """

    def __init__(self, func=np.std):
        GroupBlock.__init__(self)
        self.func = func

    def _do_use(self, inp):
        output = self._new_output()
        sp = ft.Spectrum()
        sp.wavelength = np.copy(inp.wavelength)
        sp.flux = self.func(inp.matrix(), 0)
        if len(sp.flux) != len(sp.wavelength):
            raise RuntimeError("func returned vector of length %d, but should be %d" % (len(sp.flux), len(sp.wavelength)))
        output.add_spectrum(sp)
        return output


class GB_SNR(GroupBlock):
    """Calculates the GB_SNR(lambda) = Power_signal(lambda)/Power_noise(lambda)

    **Warning** Don't use this block, empirical SNR is usually calculated **per spectrum**, as
    reinforced by EC and BLB (using IRAF procedure, which is Amplitude_RMS/std

    Args:
        continua: SpectrumList containing the continua that will be used as the "signal" level.
                    If not passed, will be calculated from the input spectra using a SB_Rubberband(True) block

    References:
        [1] https://en.wikipedia.org/wiki/Signal_averaging
    """

    # TODO I think that it is more correct to talk about "continuum" not continua

    def __init__(self, continua=None):
        GroupBlock.__init__(self)
        self.continua = continua

    def _do_use(self, inp):
        if self.continua is None:
            continua = ex.SLB_UseSpectrumBlock(sb.SB_Rubberband(True)).use(inp)
        else:
            continua = self.continua
        cont_2 = ex.SLB_UseSpectrumBlock(sb.SB_ElementWise(np.square)).use(continua)  # continua squared
        mean_cont_2 = GB_UseNumPyFunc(np.mean).use(cont_2)
        var_spectra = GB_UseNumPyFunc(np.var).use(inp)
        output = self._new_output()
        sp = ft.Spectrum()
        sp.wavelength = np.copy(inp.wavelength)
        sp.flux = mean_cont_2.spectra[0].flux/var_spectra.spectra[0].flux
        output.add_spectrum(sp)
        return output
